"""
Single Cell M
=============

The script is split into 3 commands:

* filter
* lca
* report

Which are supposed to be run in that order, since the output of **filter** is
used by **lca** and so on. However, they can be run in a out of order fashion
to experiment with different options to refine the analysis.

A separate command called *orf* can be used to report the Last Common Ancestor
for Open Read Frame (ORF) called used Prodigal.

Requirements
------------

The script is intended to be used on an already assembled single cell sample,
on which a BLAST (blastn/blastx/tblastx) or Diamond (blastx) similarity search
was performed. The output from BLAST is filtered (*filter*), the LCA (*lca*)
for each contig is searched and finally a series of reports and figures
(*report*) is produced, which directs the analysis of the sample for
contaminations and unwanted genetic material.

Exit Codes
----------

* 1 - Taxon name was not found
* 2 - Taxon name has more than 1 ID
* 3 - Taxon ID was not found in the Taxonomy
* 4 - The training set (PCA) has no contigs
* 5 - The coverage file could not be loaded
* 6 - Not enough sequences (less than 2) can be loaded.
* 7 - The taxa table was not found in the HDF5 file supplied
* 8 - Fails the DB tests on the *wf* command
* 9 - The taxonomy file cannot be found
* 10 - The HDF5 table file can't be found
"""
from __future__ import division
import sys
import argparse
import logging
import matplotlib

# Necessary before another backend may be selected
matplotlib.use('Agg')

import mgkit
import mgkit.taxon

from . import lca
from . import blast_filter
from . import workflow
from . import report
from .ui.utils import basic_ui_options

LOG = logging.getLogger(__name__)


def set_common_parser(parser):
    parser.add_argument(
        '-t',
        '--taxonomy',
        action='store',
        type=str,
        default=None,
        required=True,
        help='Taxonomy file',
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        action='store',
        default='report-output',
        help='Output directory',
    )
    parser.add_argument(
        '--log',
        action='store',
        default=None,
        help='File to log messages',
    )


def set_report_parser(parser):
    set_common_parser(parser)
    parser.add_argument(
        '-f',
        '--sequences',
        action='store',
        type=argparse.FileType('rb'),
        required=True,
        help='Fasta file',
    )
    target_taxon = parser.add_mutually_exclusive_group(required=False)
    target_taxon.add_argument(
        '-i',
        '--target-taxon-id',
        action='store',
        type=int,
        default=None,
        required=False,
        help='Target Taxon ID',
    )
    target_taxon.add_argument(
        '-n',
        '--target-taxon-name',
        action='store',
        type=str,
        default=None,
        required=False,
        help='Target Taxon name',
    )
    parser.add_argument(
        '-e',
        '--target-threshold',
        action='store',
        type=float,
        default=50.,
        required=False,
        help='Percentage of sequences to find the target taxon (0 to 100)'
    )
    parser.add_argument(
        '-l',
        '--lca-json',
        action='store',
        type=argparse.FileType('r'),
        default=None,
        required=True,
        help='LCA json file from the *lca* command',
    )
    parser.add_argument(
        '-g',
        '--rank',
        action='store',
        choices=mgkit.taxon.TAXON_RANKS,
        default='genus',
        required=False,
        help='Taxonomic rank to group taxa',
    )
    parser.add_argument(
        '-p',
        '--lineage-separator',
        action='store',
        type=str,
        default=';',
        required=False,
        help='Separator for elements of the lineage',
    )
    parser.add_argument(
        '-w',
        '--window-size',
        action='store',
        type=int,
        default=5000,
        required=False,
        help='Window size for kmer counting'
    )
    parser.add_argument(
        '-s',
        '--step',
        action='store',
        type=int,
        default=500,
        required=False,
        help='Step for kmer counting'
    )
    parser.add_argument(
        '-k',
        '--kmer-size',
        action='store',
        type=int,
        default=5,
        required=False,
        help='kmer counting size'
    )
    parser.add_argument(
        '-b',
        '--bins',
        action='store',
        type=int,
        default=20,
        required=False,
        help='Number of bins for the GC plot'
    )
    parser.add_argument(
        '-x',
        '--exclude-seq',
        action='store',
        type=argparse.FileType('rb'),
        default=None,
        required=False,
        help='Contigs to exclude from the analysis (one per line)'
    )
    parser.add_argument(
        '-r',
        '--coverage-file',
        action='store',
        type=argparse.FileType('r'),
        default=None,
        # required=True,
        help='Contig coverage file (JSON)'
    )
    parser.add_argument(
        '-c',
        '--min-cov',
        action='store',
        type=float,
        default=None,
        required=False,
        help='Minimum annotation coverage to keep the sequence, requires *-r*'
    )
    parser.add_argument(
        '-q',
        '--fastq',
        action='store_true',
        default=False,
        help='The sequences are in FASTQ format',
    )
    parser.add_argument(
        'input_file',
        nargs='*',
        help='Filtered GFF(s)',
        type=argparse.FileType('rb'),
        default=sys.stdin
    )
    parser.set_defaults(func=report.report_command)


def set_lca_contig_parser(parser):
    set_common_parser(parser)
    parser.add_argument(
        'input_file',
        nargs='*',
        help='Filtered GFF(s)',
        type=argparse.FileType('rb'),
        default=sys.stdin
    )
    parser.set_defaults(func=lca.lca_contig_command)


def parse_hdf5_arg(argument):
    file_name, table = argument.strip().split(':')
    return (file_name, table)


def set_filter_options(parser):
    parser.add_argument(
        '-d',
        '--db-type',
        action='store',
        type=str,
        default='ncbi',
        choices=('ncbi', 'uniprot', 'other', 'ncbi-full'),
        help='DB Used - *ncbi* refers to both *nt* and *nr*',
    )
    parser.add_argument(
        '-i',
        '--identity',
        action='store',
        type=float,
        default=.97,
        help='Quantile to keep out of all hits',
    )
    parser.add_argument(
        '-f',
        '--sequences',
        action='store',
        type=argparse.FileType('rb'),
        required=True,
        help='Fasta file',
    )
    parser.add_argument(
        '-e',
        '--evalue',
        action='store',
        type=float,
        default=.01,
        help='Only hits with an evalue lower than this will be kept (if found)',
    )
    parser.add_argument(
        '-a',
        '--table',
        action='store',
        type=parse_hdf5_arg,
        default=None,
        required=True,
        help='HDF5 file with gene_id to taxon_id table (e.g. taxa-table.hf5:ncbi-nucl)',
    )
    parser.add_argument(
        '-x',
        '--exclude-virus',
        action='store_true',
        default=False,
        help='Exclude hits on viruses',
    )
    parser.add_argument(
        '-r',
        '--no-taxa-filter',
        action='store_true',
        default=False,
        help='No taxa filter based on Cellular Organisms or Viruses',
    )
    parser.add_argument(
        '-q',
        '--fastq',
        action='store_true',
        default=False,
        help='The sequences are in FASTQ format',
    )
    parser.add_argument(
        '-m',
        '--max-hits',
        action='store',
        default=100,
        help='Maximum number of hits to use per contig (defaults to 100)',
    )


def set_orf_parser(parser):
    set_common_parser(parser)
    set_filter_options(parser)
    parser.add_argument(
        'input_file',
        nargs='*',
        help='Blast output(s)',
        type=argparse.FileType('rb'),
        default=sys.stdin
    )
    parser.set_defaults(func=blast_filter.orf_filter_command)


def set_filter_blast(parser):
    set_common_parser(parser)
    set_filter_options(parser)
    parser.add_argument(
        '-c',
        '--coverage',
        action='store',
        type=float,
        default=0.,
        help='Minimum percentage of the contig that needs to be covered',
    )
    parser.add_argument(
        'input_file',
        nargs='*',
        help='Blast output(s)',
        type=argparse.FileType('rb'),
        default=sys.stdin
    )
    parser.set_defaults(func=blast_filter.blast_filter_command)


def set_wf_parser(parser):

    # Blast/Diamond options
    parser.add_argument(
        '-d',
        '--db',
        action='store',
        type=str,
        default=None,
        required=True,
        help='DB to be used with blastn/blastx/diamond',
    )
    # Filter options
    parser.add_argument(
        '-b',
        '--db-type',
        action='store',
        type=str,
        default='ncbi-nucl',
        choices=('ncbi-nucl', 'uniprot', 'ncbi-prot'),
        help='''DB used, ncbi-* assumes a NCBI BLAST index, for Uniprot
            Diamond will be used if the DB ends in .dmnd''',
    )
    parser.add_argument(
        '--header-full',
        action='store_true',
        default=False,
        help='''Use is NCBI BLAST indexes were downloaded, to set the header
            type for the filter command option `-d` to `ncbi-full`''',
    )
    parser.add_argument(
        '-a',
        '--table',
        action='store',
        type=str,
        default=None,
        required=True,
        help='HDF5 file with gene_id to taxon_id table',
    )
    parser.add_argument(
        '--table-name',
        action='store',
        type=str,
        default=None,
        help='HDF5 table inside the file, defaults to --db-type',
    )
    # Other
    parser.add_argument(
        '-t',
        '--taxonomy',
        action='store',
        type=str,
        default=None,
        required=True,
        help='Taxonomy file',
    )
    parser.add_argument(
        '-w',
        '--overwrite',
        action='store_true',
        default=False,
        help='''If passed, overwrites the content of the output directories
            while the default is to skip existing directories'''
    )
    parser.add_argument(
        '-s',
        '--suffix',
        default='output',
        action='store',
        type=str,
        help="""Suffix to add to the fasta file name (after removing its
            extension and adding a hyphen) as the output directory"""
    )
    parser.add_argument(
        '--log',
        action='store',
        default=None,
        help='File to log messages',
    )
    parser.add_argument(
        '-q',
        '--fastq',
        action='store_true',
        default=False,
        help='The sequences are in FASTQ format',
    )
    parser.add_argument(
        'sequence_files',
        action='store',
        type=argparse.FileType('rb'),
        nargs='+',
        help='Sequence files to process'
    )

    parser.set_defaults(func=workflow.workflow_command)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Single Cell-M',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    subparsers = parser.add_subparsers()

    # Batch file creation
    parser_wf_contig = subparsers.add_parser(
        'wf',
        help='Makes a batch file to executes all commands'
    )
    set_wf_parser(parser_wf_contig)
    basic_ui_options(parser_wf_contig, manual=workflow.command.__doc__)

    # LCA
    parser_lca_contig = subparsers.add_parser(
        'lca',
        help='Produces a Last Common Ancestor table'
    )

    set_lca_contig_parser(parser_lca_contig)
    basic_ui_options(parser_lca_contig, manual=lca.command.__doc__)

    # Filter Blast/Diamond Output
    parser_filter_blast = subparsers.add_parser(
        'filter',
        help='Filters BLAST hits'
    )

    set_filter_blast(parser_filter_blast)
    basic_ui_options(parser_filter_blast, manual=blast_filter.command.__doc__)

    # Report
    parser_report = subparsers.add_parser(
        'report',
        help='Produces HTML reports and figures'
    )

    set_report_parser(parser_report)
    basic_ui_options(parser_report, manual=report.command.__doc__)

    # Filter prodigal aa
    parser_orf_filter = subparsers.add_parser(
        'orf',
        help='Filters BLAST hits from Open Reading Frames'
    )

    set_orf_parser(parser_orf_filter)
    basic_ui_options(parser_orf_filter, manual=blast_filter.command.__doc__)

    # for the main command
    basic_ui_options(parser, manual=__doc__)

    return parser


def main():
    "Main function"

    options = set_parser().parse_args()

    if options.log is not None:
        mgkit.logger.config_log_to_file(
            level=logging.DEBUG,
            output=options.log
        )

    mgkit.logger.config_log(options.verbose)

    # Set the minimum level of messages, but the console messages will be only
    # at the value in options.verbose
    logging.getLogger().setLevel(logging.DEBUG)

    options.func(options)



from mgkit import taxon


def get_gtdb_like_taxonomy(taxon_id, taxonomy, sep=';'):
    lineage = []
    taxa = taxon.get_lineage(
        taxonomy,
        taxon_id,
        only_ranked=True,
        with_last=True
    )
    for taxon_id in taxa:
        taxon_name = taxonomy[taxon_id]
        taxon_name = "{}_{}".format(
            taxon_name.rank[0] if taxon_name.rank != 'superkingdom' else 'k',
            taxon_name.s_name
        )
        lineage.append(taxon_name)
    return sep.join(lineage)



"""
Workflow Command
================

This command prepares a bash script that can be used to run in succession the
commands necessary to perform the analysis, from the sequence file (FASTA or
FASTQ) to the final report. The options used for each command are the mostly
conservative or the default ones and they should be customised for better
performance and analysis. If the templates here don't fit your workflow, it is
still a good idea to generate a similar one and customise the resulting bash
script.

The command produces 3 different templates based on the type of database (DB)
chosen and the software to seach for similarity between DB and sequences
provided. The DBs supported with this command are NCBI (*nt* and *nr*) and
Uniprot. The supported software is *blastn* (for NCBI *nt*), *blastx* (for NCBI
*nr*) and diamond (for Uniprot).

Options/Requirements
--------------------

Requirements
************

The script requires, besides the sequence file to analyse, a BLAST/Diamond DB
and the file containing the ID->Taxon table and a taxonomy file.

Obtain a DB
^^^^^^^^^^^

To obtain a copy of NCBI BLAST DBs, you can download all files from `NCBI FTP`_:

    * files starting with **nr** for proteins (used *blastx*)
    * files starting with **nt** for nucleotides (use *blastn*)

The files should be compressed (ending with .gz) and after using `gunzip` they
can be used. The files ending with .md5 are used to check the download was
correct.

To obtain a copy of Uniprot DBs, you need to download them at `Uniprot`_, and
there is both reviewed and unreviewed DBs:

* Reviewed: ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/uniprot_sprot.fasta.gz
* Unreviewed: ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/uniprot_trembl.fasta.gz

The DBs are in FASTA format and need to be indexed with `Diamond`_, for example::

    $ diamond makedb --in uniprot_sprot.fasta.gz -d swissprot

Will index the reviewed Uniprot DB and make an index file called swissprot.dmnd.

If you need to combine the 2 databases and index them, you can use a command
like this (in bash shell)::

    $ diamond makedb --in <(gunzip -c uniprot_sprot.fasta.gz uniprot_trembl.fasta.gz) -d uniprot

This command will create an index file called uniprot.dmnd that contains both
DBs.

Obtain the ID->Taxon file
^^^^^^^^^^^^^^^^^^^^^^^^^

You can get the correct file by runnig the included script `scm-download-taxa.sh`.
To get the correct file, you pass:

* *nucl* to get the NCBI *nt* references
* *prot* to get the NCBI *nr* references
* *uniprot* to get the Uniprot (both Reviewed and Unreviewed) references

For example, to download the NCBI *nt* references, you can run::

    $ scm-download-taxa.sh nucl

A file (HDF5 format) called *taxa-tables.hdf* will be created and it will
contain a *ncbi-nucl* table.

Taxonomy file
^^^^^^^^^^^^^

To downlad the taxonomy file, you can run the script::

    $ download-taxonomy.sh

That will create a file called *taxonomy.pickle* and this is included in MGKit,
which is installed along this package.

Options
*******

positional arguments:
  sequence_files        Sequence files to process

optional arguments:
  -h, --help            show this help message and exit
  -d DB, --db DB        DB to be used with blastn/blastx/diamond
  -b {ncbi-nucl,uniprot,ncbi-prot}, --db-type {ncbi-nucl,uniprot,ncbi-prot}
                        DB used, ncbi-* assumes a NCBI BLAST index, for
                        Uniprot Diamond will be used if the DB ends in .dmnd
  -a TABLE, --table TABLE
                        HDF5 file with gene_id to taxon_id table
  --table-name TABLE_NAME
                        HDF5 table inside the file, defaults to --db-type
  -t TAXONOMY, --taxonomy TAXONOMY
                        Taxonomy file
  -w, --overwrite       If passed, overwrites the content of the output
                        directories while the default is to skip existing
                        directories
  -s SUFFIX, --suffix SUFFIX
                        Suffix to add to the fasta file name (after removing
                        its extension and adding a hyphen) as the output
                        directory
  --log LOG             File to log messages
  -q, --fastq           The sequences are in FASTQ format
  -v, --verbose         more verbose - includes debug messages
  --quiet               less verbose - only error and critical messages
  --manual              Show the script manual
  --version             show program's version number and exit


Links
=====

.. _NCBI FTP: ftp://ftp.ncbi.nlm.nih.gov/blast/db/
.. _Diamond: https://github.com/bbuchfink/diamond
.. _Uniprot: https://www.uniprot.org/downloads
"""

import logging
import os.path
import sys
from glob import glob
from datetime import datetime
import json

import pandas
from mgkit.workflow.utils import exit_script

from .. import VERSION

LOG = logging.getLogger(__name__)


def get_sequence_similarity_cmd(seq_file, db_type, db, output_file, fastq):

    if db_type == 'ncbi-nucl':
        prog_name = 'blastn'
    elif db_type == 'ncbi-prot':
        prog_name == 'blastx'
    else:
        if db.endswith('.dmnd'):
            prog_name = 'diamond blastx'
        else:
            prog_name = 'blastx'

    if prog_name.startswith('blast'):
        cmd_line = '{prog_name} -query {seq_file} -db {db} -outfmt 6 -out {output_file}'
        if seq_file.endswith('.gz'):
            seq_file = '<(gunzip -c {})'.format(seq_file)
    else:
        cmd_line = '# tested on diamond v.9.9\n{prog_name} -q {seq_file} -d {db} -f 6 -o {output_file}'
        if not fastq:
            cmd_line += ' --more-sensitive'

    return '# Sequence similarity search\n' + cmd_line.format(
        prog_name=prog_name,
        db=db,
        output_file=output_file,
        seq_file=seq_file
    )


def get_filter_cmd(seq_file, db_type, taxonomy, taxa_table, output_dir, tab_file,
                    script_name, fastq, db_table, full_header):

    if db_type.startswith('ncbi'):
        db_type = 'ncbi-full' if full_header else 'ncbi'
    else:
        db_type = 'uniprot'

    if fastq:
        more = '-m 3 -q'
    else:
        more = ''

    cmd_line = '{script_name} filter -o {output_dir} -t {taxonomy} -d {db_type} -f {seq_file} -a {taxa_table}:{db_table} {more} {tab_file}'

    return '# Similarity search filtering\n' + cmd_line.format(
        script_name=script_name,
        taxonomy=taxonomy,
        db_type=db_type,
        seq_file=seq_file,
        taxa_table=taxa_table,
        output_dir=output_dir,
        db_table=db_table,
        tab_file=tab_file,
        more=more
    )


def get_lca_cmd(gff_file, output_dir, taxonomy, script_name):
    cmd_line = '{script_name} lca -t {taxonomy} -o {output_dir} {gff_file}'

    return '# LCA processing\n' + cmd_line.format(
        script_name=script_name,
        output_dir=output_dir,
        gff_file=gff_file,
        taxonomy=taxonomy
    )


def get_report_cmd(script_name, taxonomy, output_dir, seq_file, lca_file, cov_file, fastq):
    cmd_line = "{script_name} report -t {taxonomy} -o {output_dir} -f {seq_file} -l {lca_file} -r {cov_file}"
    if fastq:
        cmd_line += '-q'

    return '# Report generation\n' + cmd_line.format(
        script_name=script_name,
        taxonomy=taxonomy,
        output_dir=output_dir,
        seq_file=seq_file,
        lca_file=lca_file,
        cov_file=cov_file
    )


def test_blast_db(db, db_prefix, db_exts):

    db_path, db_name = os.path.split(db)
    db_files = glob(os.path.join(db_path, '{}*'.format(db_prefix)))

    if not db_name.startswith(db_prefix):
        LOG.warning("The DB passed doesn't look like a DB from NCBI")

    if set(db_file.split('.')[-1] for db_file in db_files) != db_exts:
        exit_script("The DB passed doesn't contain a recognisable DB", 8)

    return os.path.join(db_path, db_prefix)


def workflow_command(options):

    nt_exts = {'nal', 'nhd', 'nhi', 'nhr', 'nin', 'nnd', 'nni', 'nog', 'nsd', 'nsi',
                'nsq'}

    nr_exts = {'pal', 'phd', 'phi', 'phr', 'pin', 'pnd', 'pni', 'pog', 'ppd', 'ppi',
                'psd', 'psi', 'psq'}

    script_name = os.path.basename(sys.argv[0])

    # If the file name of the DB indicates that it's a diamond DB, switches to
    # using uniprot as DB type
    if options.db.endswith('.dmnd'):
        options.db_type = 'uniprot'

    # can be overid by passing the --table-name
    if options.db_type.startswith('ncbi-') and (options.table_name is None):
        # implies a BLAST DB, can be checked for consistency, to a point
        db_path = test_blast_db(
            options.db,
            'nt' if options.db_type.endswith('nucl') else 'nr',
            nt_exts if options.db_type.endswith('nucl') else nr_exts,
        )
    else:
        # too many possible cases, trusts the user to pass the correct one
        db_path = options.db

    # Tries to find the taxonomy file, exit the script otherwise
    if not os.path.exists(options.taxonomy):
        exit_script(
            'Cannot find a taxonomy file, check the file name or run the following command: download-taxonomy.sh {}'.format(options.taxonomy),
            9
        )

    # Tests the HDF5 file, suggesting the command relative to the DB used, if
    # not found
    if not os.path.exists(options.table):
        exit_script(
            'Cannot find the HDF5 table file, run the following command: scm-download-taxa.sh {}'.format(options.db_type.split('-')[-1]),
            10
        )
    #
    if options.table_name is not None:
        table_name = options.table_name
    else:
        table_name = options.db_type
    with pandas.HDFStore(options.table, mode='r') as hdf_file:
        if '/{}'.format(table_name) not in hdf_file.keys():
            exit_script(
                'Cannot find table "{}" in HDF5 file {}'.format(table_name, options.table),
                7
            )

    for sequence_file in options.sequence_files:
        sequence_file = sequence_file.name
        output_dir = "{}-{}".format(
            os.path.basename(os.path.splitext(sequence_file)[0]),
            options.suffix,
        )
        try:
            os.mkdir(output_dir)
        except OSError:
            # Skips existing directories
            if not options.overwrite:
                LOG.warning(
                    "Output directory %s for file %s exists, skipping it (pass -w if you want to overwrite it)",
                    output_dir,
                    sequence_file,
                )
                continue
        tab_file = os.path.join(
            output_dir,
            'blast-hits.tab'
        )

        commands = [
            '#!/bin/bash',
            '# Automatically generated by "{}" version {} on {}'.format(
                script_name,
                VERSION,
                datetime.now()
            )
        ]

        similarity_cmd = get_sequence_similarity_cmd(
            sequence_file,
            options.db_type,
            db_path,
            tab_file,
            options.fastq
        )
        commands.append(similarity_cmd)

        filter_cmd = get_filter_cmd(
            sequence_file,
            options.db_type,
            options.taxonomy,
            options.table,
            output_dir,
            tab_file,
            script_name,
            options.fastq,
            table_name,
            options.header_full
        )
        commands.append(filter_cmd)

        lca_cmd = get_lca_cmd(
            os.path.join(output_dir, 'blast-hits.gff'),
            output_dir,
            options.taxonomy,
            script_name
        )
        commands.append(lca_cmd)

        report_cmd = get_report_cmd(
            script_name,
            options.taxonomy,
            output_dir,
            sequence_file,
            os.path.join(output_dir, 'lca-contigs.json'),
            os.path.join(output_dir, 'contig-cov.json'),
            options.fastq
        )
        commands.append(report_cmd)

        with open(output_dir + '.bash', 'w') as output_file:
            output_file.write('\n\n'.join(commands))
            output_file.write('\n')



import logging
import json
import sys
import os.path
from datetime import datetime

from mgkit.workflow.utils import PrintManAction
from .. import VERSION


def basic_ui_options(parser, manual=''):

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-v',
        '--verbose',
        action='store_const',
        const=logging.DEBUG,
        default=logging.INFO,
        help='more verbose - includes debug messages',
        dest='verbose'
    )
    group.add_argument(
        '--quiet',
        action='store_const',
        const=logging.ERROR,
        help='less verbose - only error and critical messages',
        dest='verbose'
    )

    parser.add_argument(
        '--manual',
        action=PrintManAction,
        manual=manual
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {0}'.format(VERSION)
    )

def write_manifest(output_dir):

    json.dump(
        dict(
            script=os.path.basename(sys.argv[0]),
            command=sys.argv[1],
            version=VERSION,
            cmd_line=' '.join(sys.argv),
            datetime=str(datetime.now())
        ),
        open(
            os.path.join(
                output_dir,
                '{}-manifest.json'.format(sys.argv[1])
            ),
            'w'
        ),
    )



"""
Report Command
==============

The command will output a figure and a HMTL file with a plot/table that can be
used to inspect the kmer counts on a series of windows for each contig.

Options/Requirements
--------------------

Required
********

* taxonomy file (-t option), same in other commands
* assembly file (-f option), FASTA file with the contigs
* LCA JSON file (-l option) from the **lca** commnad

The taxonomy file is the same requirement for commands run earlier in the
pipeline, while the assembly file is the same file that was used to make the
BLAST ouput used with the **filter** command. It is expected to be a FASTA file.

The LCA (last common ancestor) file is the output of the **lca** command, and
is a JSON file that contains the information about is needed to direct the
target taxon assignment.

Optional
********

* taxonomic rank used to group the taxa by colour (-g option)
* contig coverage JSON file (-r option), from the *filter* command
* lineage separator (-s option)
* options for the kmer counting
* number of bins for the GC content histogram
* exclude contigs from analysis
* exclude contigs based on their annotation coverage
* target Taxon, either a name (e.g. bacteria, escherichia) or ID from NCBI taxonomy

The rank (defaults to *genus*) is used to colour groups of taxa, so can they be
visually identified. The ranks are the ones accepted in the taxonomy (e.g.
family, order, phylum, etc.).

The lineage separator is the character(s) used to separate the taxa the lineage
consists of. It's used in all outputs from this command.

The options for kmer counting change the behaviour of the plot by setting the
window size (-w option, defaults to 5000), step (-s option, defaults to 500),
and kmer length (-k option, defaults to 5). Lower window sizes and steps make
the contigs appear more similar, while a greater kmer size make the computation
more memory demanding. Usually a value of 4 or 5 is used, with 9 requiring
much more memory. Moreover, any contig whose length is less than the window
size specified will be excluded from the analysis.

It is also possible to exclude some contigs (-x option) from the computation by
passing a text file with the headers to exclude, as copied from the HTML report,
with one header per line.

Another option is to exclude contigs with a coverage less than the chosen
threshold (-c option, values from 0 to 100).

The target taxon is based on `NCBI taxonomy <https://www.ncbi.nlm.nih.gov/entrez/query.fcgi?db=Taxonomy>`_
and all the taxa below the target one are then marked as target. For example if
the target taxon is set to "escherichia", all species that belong to that genus
will be marked as meeting the target. It can be specified as either a name (-n
option) or as a ID (numeric, -i option). The name is not accepted if the name
is not unique in the taxonomy, in which case the script exit. Valid target taxa
are high level ones like *bacteria*, *archaea* or *viruses*.

A target taxon is automatically chosen by getting a taxon that contains at
least half of all the sequences identified (via LCA). The percentage value can
be modified by using the *-e* option, passing a value between 0 and 100, but
very high or low numbers should be avoided.

The file *contig-cov.json* (produced by the **filter** command) needs to be
passed (-r option) and it's used to put information in *PCA* plot.

HTML Reports
------------

The HTML (file **pca-plot.html**) report contains a PCA plot of the first 2
dimentions, using the contigs that are within the target taxon as training set.
The plot is linked to the a sortable table below it, which can be also be used
to select points in the plot (alternative to using the lasso selection). Any
contigs selected in the plot or table will be written appear under the table,
so it can be copied in a file (for example if in a further analysis they must
be excluded). Moreover, the colours of the lines/dots are the same as the other
plot (a PDF file) that is in the same directory.

Another HTML file is produced (**gc-plot.html**) that contains an histogram of
the GC content of the contigs analysed. When hovering over one of the bars,
some information is displayed as tooltip and the headers of the contigs
included in that bin is displayed in a section underneat the histogram.
"""
from __future__ import division

import os
import logging
import json
from collections import Counter

import pandas
from sklearn.decomposition import PCA
from sklearn.cluster import MeanShift
import distance

from mgkit.workflow.utils import exit_script
from mgkit.io import fasta, fastq
from mgkit import taxon
from mgkit.utils import sequence

from . import static
from . import interactive
from ..ui.utils import write_manifest

LOG = logging.getLogger(__name__)


def training_sets(sequences, lca_dict, taxonomy, target_taxon):
    seq_ids_training = list(
        seq_id
        for seq_id in sequences
        if (seq_id in lca_dict) and taxonomy.is_ancestor(
            lca_dict[seq_id]['taxon_id'], int(target_taxon)
        )
    )

    seq_ids_unknown = list(
        set(sequences) - set(seq_ids_training)
    )

    return seq_ids_training, seq_ids_unknown


def kmer_counting(sequences, kmer_size, window_size, step):
    LOG.info(
        "Kmer counting (kmer %i, window %i, step %i) for %i sequences",
        kmer_size,
        window_size,
        step,
        len(sequences)
    )

    matrix = sequence.signatures_matrix(
        sequences.items(),
        k_size=kmer_size,
        w_size=window_size,
        step=step
    )

    matrix.sort_index(level=0, inplace=True)

    return matrix


def pca_transform(matrix, seq_ids_training, seq_ids_unknown):

    LOG.info("Performing PCA transform")

    pca = PCA(n_components=3, whiten=False)

    pca.fit(matrix.loc[seq_ids_training])

    tsf_training = pandas.DataFrame(
        pca.transform(matrix.loc[seq_ids_training]),
        index=matrix.loc[seq_ids_training].index,
        columns=['X', 'Y', 'Z']
    )
    if len(seq_ids_unknown) > 0:
        tsf_unknown = pandas.DataFrame(
            pca.transform(matrix.loc[seq_ids_unknown]),
            index=matrix.loc[seq_ids_unknown].index,
            columns=['X', 'Y', 'Z']
        )
    else:
        LOG.warning(
            "The unknown set contains no contigs. All fall within the " +
            "specified target"
        )
        tsf_unknown = None

    LOG.info("Explained variance %r", pca.explained_variance_ratio_ * 100)

    return tsf_training, tsf_unknown, pca.explained_variance_ratio_ * 100


def ms_cluster(matrix):

    LOG.info("Performing MeanShift clustering")

    pca = PCA(n_components=3, whiten=False)

    centroids = matrix.groupby(level=0).mean()

    transformed = pca.fit_transform(centroids)

    ms = MeanShift()

    return dict(
        zip(
            centroids.index,
            ms.fit_predict(transformed),
        )
    )


def count_sequences(lca_dict, taxonomy, rank):
    counts = Counter(
        d['taxon_id'] if rank is None else taxonomy.get_ranked_taxon(d['taxon_id'], rank).taxon_id
        for seq_id, d in lca_dict.items()
        if d['taxon_id'] is not None
    )
    counts = pandas.Series(counts).sort_values(ascending=False)
    return counts.divide(counts.sum()) * 100


def find_target_taxon(lca_dict, taxonomy, threshold=50.):
    LOG.info("Finding target taxon (>= %.2f of contigs)", threshold)
    counts = count_sequences(lca_dict, taxonomy, None)
    if counts.iloc[0] >= threshold:
        return counts.index[0]

    for rank in reversed(taxon.TAXON_RANKS):
        counts = count_sequences(lca_dict, taxonomy, rank)
        if counts.iloc[0] >= threshold:
            return counts.index[0]


def report_command(options):
    try:
        os.mkdir(options.output_dir)
    except OSError:
        pass
        # Add exit if directory exists?

    try:
        contig_coverage = json.load(options.coverage_file)
    except (ValueError, AttributeError):
        if options.min_cov is not None:
            exit_script(
                "To use the coverage filter you need to pass the " +
                "coverage file (JSON)", 5
            )
        else:
            contig_coverage = {}

    taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    lca_dict = json.load(options.lca_json)

    if (options.target_taxon_id is None) and (options.target_taxon_name is not None):
        try:
            target_taxon = taxonomy.find_by_name(options.target_taxon_name)
        except KeyError:
            names = sorted(
                distance.ifast_comp(
                    options.target_taxon_name,
                    (x.s_name for x in taxonomy)
                )
            )[:10]
            exit_script(
                "Taxon name '{}' not found, possible alternatives: {}".format(
                    options.target_taxon_name,
                    ', '.join(x[1] for x in names)
                ),
                1
            )
        if len(target_taxon) > 1:
            exit_script(
                "Taxon name '{}' has multiple taxa ID".format(
                    options.target_taxon_name
                ),
                2
            )
        target_taxon = target_taxon[0]
    elif (options.target_taxon_id is not None) and (options.target_taxon_name is None):
        if options.target_taxon_id not in taxonomy:
            exit_script(
                "Taxon ID '{}' was not found in the Taxonomy".format(
                    options.target_taxon_id
                ),
                3
            )
        target_taxon = options.target_taxon_id
    else:
        target_taxon = find_target_taxon(
            lca_dict,
            taxonomy,
            options.target_threshold
        )

    LOG.info(
        "The target taxon is '%s' (rank '%s')",
        taxonomy[target_taxon].s_name,
        taxonomy[target_taxon].rank
    )

    if options.exclude_seq is None:
        exclude_seqs = set()
    else:
        exclude_seqs = set(
            line.strip()
            for line in options.exclude_seq
        )

    if options.min_cov is not None:
        LOG.info(
            "Accepting only contigs with %.2f%% of coverage",
            options.min_cov
        )
        for seq_id, cov in contig_coverage.items():
            if cov < options.min_cov:
                exclude_seqs.add(seq_id)

    load_sequences_func = fastq.load_fastq_rename if options.fastq else fasta.load_fasta_rename
    sequences = dict(
        sequence[:2]
        for sequence in load_sequences_func(options.sequences)
        if (len(sequence[1]) >= options.window_size) and (sequence[0] not in exclude_seqs)
    )

    if len(sequences) < 2:
        exit_script(
            "Not enough ({}) sequences could be loaded (min {} in length)".format(
                len(sequences),
                options.window_size
            ),
            6
        )

    seq_lengths = static.group_seq_lengths(
        sequences,
        lca_dict,
        taxonomy,
        options.rank
    )

    anc_colors = static.lca_contig_count_and_length(
        seq_lengths,
        os.path.join(options.output_dir, 'length-abundance-figure.pdf'),
        taxonomy,
        options.lineage_separator,
        fontsize=14
    )

    # Interactive plot

    seq_colors = interactive.contig_colors(
        taxonomy,
        anc_colors,
        sequences,
        lca_dict,
        rank=options.rank
    )

    seq_ids_training, seq_ids_unknown = training_sets(
        sequences,
        lca_dict,
        taxonomy,
        target_taxon
    )

    if len(seq_ids_training) == 0:
        exit_script(
            "The training set contains no contigs: a higher taxon should " +
            "be used as target", 4
        )

    matrix = kmer_counting(
        sequences,
        options.kmer_size,
        options.window_size,
        options.step
    )

    tsf_training, tsf_unknown, explained_variance = pca_transform(
        matrix,
        seq_ids_training,
        seq_ids_unknown
    )

    clustered = ms_cluster(matrix)

    dataframes = interactive.create_bokeh_data_source(
        [tsf_training] if tsf_unknown is None else [tsf_training, tsf_unknown],
        lca_dict,
        seq_colors,
        taxonomy,
        seq_ids_training,
        sequences,
        options.lineage_separator,
        contig_coverage,
        clustered
    )

    dataframes.to_csv(
        os.path.join(options.output_dir, 'pca_data.csv')
    )

    pca_plot_x_y, source = interactive.prepare_plot(
        dataframes,
        seq_colors,
        seq_ids_training,
        title='Components 1 and 2',
        plot_height=1024,
        plot_width=1024,
        x_label='PCA 1 ({:.2f}%)'.format(explained_variance[0]),
        y_label='PCA 2 ({:.2f}%)'.format(explained_variance[1]),
        x_col='X',
        y_col='Y',
    )
    pca_plot_x_z, source = interactive.prepare_plot(
        dataframes,
        seq_colors,
        seq_ids_training,
        title='Components 1 and 3',
        plot_height=1024 // 2,
        plot_width=1024 // 2,
        x_label='PCA 1 ({:.2f}%)'.format(explained_variance[0]),
        y_label='PCA 3 ({:.2f}%)'.format(explained_variance[2]),
        x_col='X',
        y_col='Z',
        datasource=source,
        x_range=pca_plot_x_y.x_range
    )
    pca_plot_y_z, source = interactive.prepare_plot(
        dataframes,
        seq_colors,
        seq_ids_training,
        title='Components 2 and 3',
        plot_height=1024 // 2,
        plot_width=1024 // 2,
        x_label='PCA 2 ({:.2f}%)'.format(explained_variance[1]),
        y_label='PCA 3 ({:.2f}%)'.format(explained_variance[2]),
        x_col='Y',
        y_col='Z',
        datasource=source,
        x_range=pca_plot_x_y.y_range,
        y_range=pca_plot_x_z.y_range
    )

    pca_plot_x_y.x_range = pca_plot_x_z.x_range
    pca_plot_x_y.y_range = pca_plot_y_z.x_range
    pca_plot_x_z.y_range = pca_plot_x_z.y_range

    interactive.output_html_pca(
        source,
        pca_plot_x_y,
        [pca_plot_x_z, pca_plot_y_z],
        os.path.join(options.output_dir, 'pca-plot.html')
    )

    gc_counts = interactive.prepare_gc_data(
        sequences,
        options.bins
    )

    interactive.output_html_gc_content(
        gc_counts,
        os.path.join(options.output_dir, 'gc-plot.html')
    )

    write_manifest(options.output_dir)



"""
Blast Filter Command
====================

This commmand will filter output(s) from BLAST, used with the option
*-outfmt 6*, which outputs a tab delimited file. The result of this command is
a GFF file (blast-hits.gff) with the filtered hits in the output directory.
Additionally, 2 more files are produced with the bp covered in annotations:

    * contig-cov.json
    * contig-cov.tsv

The first is used with the *report* command if contigs with a low coverage of
of annotations (-c and -r options there). The second contains is a tab
separated file with 3 columns:

    * contig header
    * contig length
    * coverage

Options/Requirements
--------------------

Requirements
************

Besides the output file(s) from BLAST, the script requires a taxonomy file and
another that contains the mappings from the database IDs to the taxonomic IDs.

The taxonomy can be obtained using the `MGKit`_ script **download-taxonomy.sh**,
which is installed as a requirement by this package, preferably using::

    $ download-taxonomy.sh taxonomy.msgpack

Passing the **.msgpack** extension results in a different format, which is
faster to load than the default one, while the rest of the name can be changed.
This is passed to the script via the **-t** option, wihch is required.

The second file that is required is the mapping of database IDs to taxonomic
IDs. This file is dependent on the database used with BLAST and by default
the headers are expected to be in EMBL/NCBI format, with elements separated by
pipe ("|") characters. `MGKit`_ provides 3 different scripts for downloading
such information, in a way that can be used by this command.

Two scipts, *download-uniprot-taxa.sh* and *download-ncbi-taxa.sh* download the
necessary files, from Uniprot and NCBI respectively, while the third,
*taxon-utils* process the files from the previous comamnds to a HDF5 file, that
is required by this command.

...

Which will download the mappings from NCBI *nt* to taxonomic IDs and can be
passed to the **-a** option of this command.

Another requirement is the FASTA file with the assembly, which is used to
filter contigs whose hits cover less than the specified percentage of the
contig's length (disabled by default).

Optional
********

Several parameter are set to default value that can be changed to suit
different type of analysis. A brief summary:

* -o output directory to store files
* -d type of DB used, which defaults to NCBI blast DBs, in particular

    * ncbi (default): NCBI blast DBs download from NCBI, contains only the
        versioned ID
    * uniprot: a '|' separated header, used by Uniprot
    * other: the entire header is used as ID for the taxonomy
    * ncbi-full: similar to *uniprot*, but specific to NCBI

* -i identity quantile to use for filtering sequences, default to 97% of the
    sequence identity distribution
* -f Fasta file with the sequences used
* -a HDF5 file, followed by a colon and the table to be used, (e.g.
    taxa.hdf:uniprot)
* -e minimum E-value to accept a BLAST hit
* -x exclude viral hits from analysis
* -c minimum coverage
* -m maximum number of hits to use for each contig

The minimum coverage options (-c) is used to exclude contigs where the hits from
BLAST (after filtering) cover less than a certain threshold. The default
behaviour is to keep all contigs.

ORF Filter and LCA
==================

Another command, **orf** allows to get the last common ancestor (LCA) of genes
called using prodigal. This gives more information on the accuracy of the LCA
call on the contigs. Most parameters are the same as the **filter** command.

Files produced:

    * orf-hits.gff
    * orf-hits.tsv

The GFF file contains the LCA information for each gene and the coordinates it
was called on. The same information is available in the TSV file.

Links
=====

.. _MGKit: https://bitbucket.org/setsuna80/mgkit
"""

from __future__ import division

import os
import itertools
import functools
import logging
import json

import pandas
import progressbar

from mgkit import taxon
from mgkit.io import blast, gff, fasta, fastq
from mgkit.io.utils import group_tuples_by_key, open_file
from mgkit.filter import gff as filter_gff
from mgkit.utils.dictionary import HDFDict
from mgkit.utils.common import ranges_length

from ..lca.command import lca_annotations
from ..ui.utils import write_manifest

LOG = logging.getLogger(__name__)


def filter_sequence(iterator, id_quantile, evalue):

    for annotations in iterator:
        annotations = [
            annotation
            for annotation in annotations
            if annotation.get_attr('evalue', float) < evalue
        ]
        identity_threshold = pandas.Series(
            annotation.get_attr('identity', float) for annotation in annotations
        ).quantile(id_quantile)

        yield [
            annotation
            for annotation in annotations
            if annotation.get_attr('identity', float) >= identity_threshold
        ]


def add_taxa_ids(iterator, taxa_dict, taxonomy, name_func, exclude_virus, max_hits=100, no_taxa_filter=False):

    iterator = blast.parse_blast_tab(
        iterator,
        ret_col=(0, 1, 2, 6, 7, 8, 9, 10, 11),
        value_funcs = (
            str,
            name_func,
            float,
            int,
            int,
            int,
            int,
            float,
            float
        )
    )

    iterator = group_tuples_by_key(
        (seq_hit for key, seq_hit in iterator),
        key_func=lambda x: x[0]
    )

    try:
        taxa_dict = HDFDict(taxa_dict[0], taxa_dict[1])
    except ValueError as e:
        utils.exit_script(
            str(e),
            7
        )

    include_taxa = (
        taxon.CELLULAR_ORGANISMS,
        taxon.VIRUS
    )
    if exclude_virus:
        include_taxa = include_taxa[0]

    bar = progressbar.ProgressBar(max_value=progressbar.UnknownLength)

    for seq_hits in bar(iterator):
        if len(seq_hits) > max_hits:
            seq_hits = seq_hits[:max_hits]

        annotations = []

        for seq_hit in seq_hits:
            try:
                gene_id = seq_hit[1]
            except IndexError:
                LOG.critical('Cannot find gene_id')
                continue

            try:
                taxon_id = taxa_dict[gene_id]
            except KeyError:
                LOG.debug("No taxon_id found for gene_id (%s)", gene_id)
                continue

            if (not no_taxa_filter) and (not taxonomy.is_ancestor(taxon_id, include_taxa)):
                continue

            annotation = gff.from_nuc_blast(
                seq_hit,
                'TEST',
                feat_type='HIT',
                taxon_id=taxon_id,
            )

            annotations.append(annotation)

        yield annotations


def get_choose_func():
    choose_func = lambda a1, a2: min(
        a1,
        a2,
        key=lambda el: tuple(
            [
                el.get_attr('identity', float),
                el.bitscore,
                len(el)
            ]
        )
    )

    choose_func = functools.partial(
        filter_gff.choose_annotation,
        overlap=100,
        choose_func=choose_func
    )

    return choose_func


def filter_overlap(iterator, gff_file, min_cov, sequences):

    choose_func = get_choose_func()

    contig_coverage = {}

    for annotations in iterator:

        annotations = gff.group_annotations(
            annotations,
            key_func=lambda x: (x.seq_id, x.strand)
        ).values()

        filtered_annotations = []

        for strand_annotations in annotations:
            filtered = filter_gff.filter_annotations(
                strand_annotations,
                choose_func=choose_func,
                sort_func=lambda x: x.bitscore,
                reverse=True
            )

            filtered_annotations.extend(filtered)

        # Needed if no annotations are kept, the following
        # filtered_annotations[0].seq_id
        # raises an IndexError exception
        if not filtered_annotations:
            continue

        # Calculate the coverage in percentage of the length
        coverage = ranges_length(
            gff.elongate_annotations(filtered_annotations)
        ) / sequences[filtered_annotations[0].seq_id] * 100

        contig_coverage[filtered_annotations[0].seq_id] = coverage

        # Check if minimum coverage is satisfied
        if coverage < min_cov:
            continue

        for annotation in filtered_annotations:
            annotation.to_file(gff_file)

    for seq_id in set(sequences) - set(contig_coverage):
        contig_coverage[seq_id] = 0.

    return contig_coverage


def blast_filter_command(options):

    try:
        os.mkdir(options.output_dir)
    except OSError:
        pass
        # Add exit if directory exists?

    if options.db_type == 'ncbi-full':
        name_func = lambda x: x.split('|')[3].split('.')[0]
    elif options.db_type == 'uniprot':
        name_func = lambda x: x.split('|')[1]
    elif options.db_type == 'other':
        name_func = lambda x: x
    elif options.db_type == 'ncbi':
        name_func = lambda x: x.split('.')[0]

    taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    if options.no_taxa_filter:
        LOG.info('No taxa will be filtered')

    gff_file = open_file(
        os.path.join(options.output_dir, 'blast-hits.gff'),
        'w'
    )

    LOG.info("Minimum contig coverage required %d%%", options.coverage)

    iterator = itertools.chain(*options.input_file)
    iterator = add_taxa_ids(
        iterator,
        options.table,
        taxonomy,
        name_func,
        options.exclude_virus,
        max_hits=options.max_hits,
        no_taxa_filter=options.no_taxa_filter
    )
    iterator = filter_sequence(
        iterator,
        options.identity,
        options.evalue
    )
    load_sequences_func = fastq.load_fastq_rename if options.fastq else fasta.load_fasta_rename
    sequences = {
        # The third argument in the tuple is only present if a FASTQ file is
        # used
        sequence[0]: len(sequence[1])
        for sequence in load_sequences_func(options.sequences)
    }
    contig_coverage = filter_overlap(iterator, gff_file, options.coverage, sequences)

    json.dump(
        contig_coverage,
        open(os.path.join(options.output_dir, 'contig-cov.json'), 'w')
    )

    with open(os.path.join(options.output_dir, 'contig-cov.tsv'), 'w') as f:
        for seq_id, cov in contig_coverage.items():
            f.write("{}\t{}\t{}\n".format(seq_id, sequences[seq_id], cov))

    write_manifest(options.output_dir)

def group_prodigal_aa(file_handle):
    return {
        d['prod_seq_id']: d
        for d in fasta.load_fasta_prodigal(file_handle)
    }


def orf_filter_command(options):

    try:
        os.mkdir(options.output_dir)
    except OSError:
        pass
        # Add exit if directory exists?

    if options.db_type == 'ncbi':
        name_func = lambda x: x.split('|')[3].split('.')[0]
    elif options.db_type == 'uniprot':
        name_func = lambda x: x.split('|')[1]
    elif options.db_type == 'other':
        name_func = lambda x: x

    taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    gff_file = open_file(
        os.path.join(options.output_dir, 'orf-hits.gff'),
        'w'
    )
    LOG.info("Writing a GFF file to (%s)", gff_file.name)

    tsv_file = open_file(
        os.path.join(options.output_dir, 'orf-hits.tsv'),
        'w'
    )

    LOG.info("Writing LCA information to (%s)", tsv_file.name)

    iterator = itertools.chain(*options.input_file)
    iterator = add_taxa_ids(
        iterator,
        options.table,
        taxonomy,
        options.exclude_virus
    )
    iterator = filter_sequence(
        iterator,
        options.identity,
        name_func,
        options.evalue
    )

    aa_seqs = group_prodigal_aa(options.sequences)

    tsv_file.write('seq_id\tindex\tlineage\tstart\tend\tstrand\ttaxon_id\tbest_hit\tgene_ids\n')

    for annotations in iterator:
        if not annotations:
            continue
        lca_id, lineage = lca_annotations(taxonomy, annotations)
        prod_seq_id = annotations[0].seq_id

        info = aa_seqs[prod_seq_id]

        tsv_file.write(
            '{seq_id}\t{idx}\t{lineage}\t{start}\t{end}\t{strand}\t{taxon_id}\t{best_hit}\t{gene_ids}\n'.format(
                taxon_id=lca_id,
                lineage=lineage,
                gene_ids=','.join(
                    set(annotation.gene_id for annotation in annotations)
                ),
                best_hit=max(annotations, key=lambda x: x.bitscore).gene_id,
                **info
            )
        )

        gff.Annotation(
            source='PRODIGAL',
            feat_type='CDS',
            taxon_id=lca_id,
            lineage=lineage,
            gene_ids=','.join(
                set(annotation.gene_id for annotation in annotations)
            ),
            best_hit=max(annotations, key=lambda x: x.bitscore).gene_id,
            **info
        ).to_file(gff_file)

    write_manifest(options.output_dir)



"""
LCA Command
============

Take the output of the **filter** command and finds the Last Common Ancestor
for each sequence in the assembly. The result of this command is 2 files with
the LCA information (lca-contigs.tsv and lca-contigs.json), in the output
directory.

Additionally, a file (lca-contigs.krona) is written. This can be passed to the
*ktImportText* script installed with Krona to produce an HTML file with a Krona
graph. Example::

    $ ktImportText -q -o lca-contigs.html lca-contigs.krona

Requirements
------------

The requirements are a taxonomy file (-t option) which takes the same taxonomy
file previously used, in addition to the GFF produced by the **filter**
command. Optionally an output directory can be specified.
"""
import os
import json
import logging

from mgkit import taxon
from mgkit.io import gff
from mgkit.io.utils import open_file

from ..ui.utils import write_manifest

LOG = logging.getLogger(__name__)


def lca_annotations(taxonomy, seq_ann):
    taxon_ids = (a.taxon_id for a in seq_ann)
    try:
        lca_id = taxon.last_common_ancestor_multiple(taxonomy, taxon_ids)
    except taxon.NoLcaFound:
        lca_id = None

    if lca_id is None:
        lineage = '',
    else:
        lineage = ';'.join(
            taxon.get_lineage(
                taxonomy,
                lca_id,
                names=True,
                with_last=True,
                only_ranked=True
            )
        )

    return lca_id, lineage


def lca_contig_command(options):
    try:
        os.mkdir(options.output_dir)
    except OSError:
        pass
        # Add exit if directory exists?

    taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    iterator = gff.parse_gff_files(options.input_file)
    iterator = gff.group_annotations_sorted(
        iterator,
        key_func=lambda x: x.seq_id
    )

    lca_dict = {}

    for seq_ann in iterator:
        lca_id, lineage = lca_annotations(taxonomy, seq_ann)

        lca_dict[seq_ann[0].seq_id] = {
            'taxon_id': lca_id,
            'lineage': lineage
        }

    json.dump(
        lca_dict,
        open(
            os.path.join(options.output_dir, 'lca-contigs.json'),
            'w'
        ),
        indent=4
    )

    csv_file = open(
        os.path.join(options.output_dir, 'lca-contigs.tsv'),
        'w'
    )

    krona_file = open(
        os.path.join(options.output_dir, 'lca-contigs.krona'),
        'w'
    )

    for seq_id, value in lca_dict.items():
        csv_file.write(
            "{seq_id}\t{taxon_id}\t{lineage}\n".format(seq_id=seq_id, **value)
        )

        if value['taxon_id'] is None:
            krona_file.write('No LCA\n')
        else:
            krona_file.write(
                "\t".join(
                    taxon.get_lineage(
                        taxonomy,
                        value['taxon_id'],
                        names=True,
                        with_last=True,
                        only_ranked=True
                    )
                ) + '\n'
            )

    write_manifest(options.output_dir)



import itertools
import logging
import seaborn
import numpy

import mgkit.plots

from .common import get_gtdb_like_taxonomy

seaborn.set_style('whitegrid')

LOG = logging.getLogger(__name__)


def group_seq_lengths(seqs, lca_dict, taxonomy, rank):
    LOG.debug("Grouping sequences lengths")
    seq_lengths = {}
    for seq_id, seq in seqs.items():
        if seq_id in lca_dict:
            taxon_id = lca_dict[seq_id]['taxon_id']
        else:
            taxon_id = None

        if taxon_id is None:
            anc_id = None
        else:
            anc_id = taxonomy.get_ranked_taxon(taxon_id, rank=rank).taxon_id

        if anc_id not in seq_lengths:
            seq_lengths[anc_id] = {}

        try:
            seq_lengths[anc_id][taxon_id].append(len(seq))
        except KeyError:
            seq_lengths[anc_id][taxon_id] = [len(seq)]

    return {
        anc_id: {
            taxon_id: numpy.array(values)
            for taxon_id, values in seq_data.items()
        }
        for anc_id, seq_data in seq_lengths.items()
    }


def lca_contig_count_and_length(seq_lengths, file_name, taxonomy, sep, fontsize=14, unknown_col='#999999'):

    LOG.info(
        "Outputting box/bar plot for contig length and abundance to (%s)",
        file_name
    )

    anc_order = sorted(
        (
            (anc_id, numpy.concatenate(list(seq_data.values())).mean())
            for anc_id, seq_data in seq_lengths.items()
        ),
        key=lambda x: x[1]
    )
    anc_order = [anc_id for anc_id, value in anc_order]
    anc_colors = {
        taxon_id: unknown_col if taxon_id is None else mgkit.plots.float_to_hex_color(*color)
        for taxon_id, color in zip(anc_order, seaborn.color_palette('hls', len(anc_order)))
    }

    plot_order = []
    for anc_id in anc_order:
        for taxon_id, lengths in seq_lengths[anc_id].items():
            plot_order.append(
                (anc_id, taxon_id)
            )

    n_rows = len(plot_order)
    fig, gs = mgkit.plots.get_grid_figure(
        1,
        2,
        figsize=(15, .25 * n_rows),
        width_ratios=[1, .75],
        wspace=.0
    )

    axb = fig.add_subplot(gs[0])

    data = [
        seq_lengths[anc_id][taxon_id]
        for anc_id, taxon_id in plot_order
    ]

    plot_data = axb.boxplot(data, vert=False, patch_artist=True, widths=0.6)

    for index, (anc_id, taxon_id) in enumerate(plot_order):
        plot_data['boxes'][index].set(
            edgecolor='black',
            facecolor=anc_colors[anc_id],
            lw=0.5
        )
        plot_data['medians'][index].set_color('black')
        plot_data['whiskers'][index].set(color='black', linestyle='--', lw=1)
        plot_data['whiskers'][index+len(plot_order)].set(
            color='black',
            linestyle='--',
            lw=1
        )
        plot_data['caps'][index].set(color='black', ls='-')

        axb.scatter(
            seq_lengths[anc_id][taxon_id],
            numpy.ones(len(seq_lengths[anc_id][taxon_id])) + index,
            color=anc_colors[anc_id],
            s=50
        )

    axb.invert_xaxis()

    axb.grid(True, which='major', axis='both')
    minor_tick = (axb.get_xticks(minor=False)[1] - axb.get_xticks(minor=False)[0]) // 2
    axb.set_xticks(
        numpy.arange(0, axb.get_xlim()[0], minor_tick),
        minor=True
    )
    axb.grid(True, which='minor', axis='both', linestyle='--')

    # _ = axb.set_xlim(right=c_df.min().min() - minor_tick)
    _ = axb.set_ylim(bottom=0, top=len(plot_order) + 0.5)

    axb.set_title('Contig Length', fontsize=fontsize)
    axb.set_xlabel('Length in bp', fontsize=fontsize)

    axc = fig.add_subplot(gs[1], sharey=axb)

    axc.barh(
        numpy.arange(1., len(plot_order) + 1., 1),
        [
            len(seq_lengths[anc_id][taxon_id])
            for anc_id, taxon_id in plot_order
        ],
        height=0.8,
        color=[anc_colors[anc_id] for anc_id, taxon_id in plot_order],
        lw=0.5
    )
    _ = axc.set_ylim(bottom=0.5, top=len(plot_order) + 0.5)

    group_labels = []
    for anc_id in anc_order:
        if anc_id is None:
            lineage = 'Unknown'
        else:
            lineage = get_gtdb_like_taxonomy(anc_id, taxonomy, sep=sep)
        group_labels.append(lineage)

    mgkit.plots.heatmap.grouped_spine(
        [seq_lengths[anc_id].keys() for anc_id in anc_order],
        group_labels,
        axc,
        start=0.5,
        spine_opts=dict(fontsize=fontsize)
    )

    axb.yaxis.set_visible(False)

    axc.set_title('Number of Contigs Assigned', fontsize=fontsize)
    axc.set_xlabel('Count', fontsize=fontsize)

    for text in axb.get_xticklabels() + axc.get_xticklabels():
        text.set_fontsize(fontsize * .75)

    fig.savefig(file_name, bbox_inches='tight')

    return anc_colors


# not used at the moment
def contig_decomposition_plot(tsf_training, tsf_unknown, seq_colors, train_alpha=0.25, fname='test-pca.pdf', include_training=True):
    fig, gs = mgkit.plots.get_grid_figure(3, 1, figsize=(7.5, 7.5 * 3))
    for gs_index, (column1, column2) in enumerate(itertools.combinations(tsf_training.columns, 2)):
        ax = fig.add_subplot(gs[gs_index])
        if include_training:
            for seq_id in tsf_training.index.get_level_values(0).unique():
                ax.plot(
                    tsf_training.loc[seq_id, column1],
                    tsf_training.loc[seq_id, column2],
                    color=seq_colors[seq_id],
                    alpha=train_alpha,
                    marker='o',
                    lw=0.75
                )
        for seq_id in tsf_unknown.index.get_level_values(0).unique():
            ax.plot(
                tsf_unknown.loc[seq_id, column1],
                tsf_unknown.loc[seq_id, column2],
                color=seq_colors[seq_id],
                alpha=1.,
                marker='o',
                lw=1.5
            )
        ax.set_xlabel(column1)
        ax.set_ylabel(column2)

        fig.savefig(fname, bbox_inches='tight')



import pandas
import logging

import bokeh
import bokeh.models
import bokeh.models.tools
from bokeh.models import HoverTool, BoxZoomTool, CustomJS
from bokeh.layouts import gridplot, column

from bokeh.resources import CDN
from bokeh.embed import file_html
import bokeh.plotting
from bokeh.models.widgets import DataTable, DateFormatter, TableColumn, TextInput, Div
from bokeh.models.sources import ColumnDataSource

import mgkit.plots
import mgkit.utils.sequence

from .common import get_gtdb_like_taxonomy


LOG = logging.getLogger(__name__)


def contig_colors(taxonomy, anc_colors, sequences, lca_dict, rank, unknown_col='#999999'):
    seq_colors = {}
    for seq_id, value in lca_dict.items():
        if value['taxon_id'] is None:
            continue
        anc_id = taxonomy.get_ranked_taxon(
            value['taxon_id'],
            rank=rank
        ).taxon_id

        try:
            seq_colors[seq_id] = anc_colors[anc_id]
        except KeyError:
            continue

    seq_colors.update(
        {
            seq_id: unknown_col
            for seq_id in set(sequences) - set(seq_colors)
        }
    )
    return seq_colors


def create_bokeh_data_source(dataframes, lca_dict, seq_colors, taxonomy, seq_ids_training, seqs, sep, contig_coverage, clustered, training_alpha=0.25):
    LOG.info("Create Bokeh Data Source")

    dataframes = pandas.concat(dataframes)
    lineages = {
        (seq_id, index): 'Unknown' if (seq_id not in lca_dict) or (lca_dict[seq_id]['taxon_id'] is None) else get_gtdb_like_taxonomy(lca_dict[seq_id]['taxon_id'], taxonomy, sep=sep)
        for seq_id, index in dataframes.index
    }
    dataframes['lineage'] = pandas.Series(lineages)
    dataframes['alpha'] = pandas.Series(
        {
            (seq_id, index): training_alpha if seq_id in seq_ids_training else 1.
            for seq_id, index in dataframes.index
        }
    )
    dataframes['target_taxon'] = pandas.Series(
        {
            (seq_id, index): seq_id in seq_ids_training
            for seq_id, index in dataframes.index
        }
    )
    dataframes['seq_id'] = pandas.Series(dataframes.index.get_level_values(0), index=dataframes.index)

    dataframes['color'] = pandas.Series(
        {
            (seq_id, index): seq_colors[seq_id]
            for seq_id, index in dataframes.index
        }
    )
    dataframes['length'] = pandas.Series(
        {
            (seq_id, index): len(seqs[seq_id])
            for seq_id, index in dataframes.index
        }
    )
    dataframes['coverage'] = pandas.Series(
        {
            (seq_id, index): contig_coverage.get(seq_id, 0)
            for seq_id, index in dataframes.index
        }
    )
    dataframes['cluster'] = pandas.Series(
        {
            (seq_id, index): clustered[seq_id]
            for seq_id, index in dataframes.index
        }
    )

    LOG.debug("Returning dataframe")

    return dataframes.sort_index(level=0).sort_values(by='target_taxon', ascending=False)


def prepare_plot(dataframes, seq_colors, seq_ids_training, title='', plot_width=400, plot_height=400, x_range=None, y_range=None, x_label='', y_label='', x_col='X', y_col='Y', datasource=None):
    LOG.info("Prepare PCA plot for %s and %s", x_col, y_col)

    if datasource is None:
        datasource = ColumnDataSource(dataframes)
    figure = bokeh.plotting.figure(
        plot_width=plot_width,
        plot_height=plot_height,
        tools=','.join(
            [
                "pan",
                "box_zoom",
                "save",
                "reset",
                "tap",
                "lasso_select",
                "box_select",
                # "resize", # doesn't work with the layout
                "wheel_zoom"
            ]
        ),
        title=title,
        x_range=x_range,
        y_range=y_range
    )

    LOG.debug("Drawing Lines")

    lines_x = []
    lines_y = []
    lines_color = []
    lines_alpha = []
    for seq_id in dataframes.index.get_level_values(0).unique():
        # Using loc[(seq_id,), x_col] triggers a PerformanceWarning from pandas
        # loc[seq_id, x_col] can't be used because the index is a MultiIndex
        lines_x.append(dataframes.loc[seq_id].loc[:, x_col].values)
        lines_y.append(dataframes.loc[seq_id].loc[:, y_col].values)
        lines_color.append(seq_colors[seq_id])
        lines_alpha.append(0.25 if seq_id in seq_ids_training else 1.)
    line_source = ColumnDataSource({'X': lines_x, 'Y': lines_y, 'color': lines_color, 'alpha': lines_alpha})
    ml = bokeh.models.MultiLine(xs='X', ys='Y', line_color='color', line_alpha='alpha')
    figure.add_glyph(source_or_glyph=line_source, glyph=ml)

    LOG.debug("Drawing Circles")

    c = bokeh.models.Circle(x=x_col, y=y_col, fill_color='color', fill_alpha=1., line_color=None, radius=1.5)
    c_render = figure.add_glyph(source_or_glyph=datasource, glyph=c)
    hover = HoverTool(
        renderers=[c_render],
        tooltips=[
            ("LCA", "@lineage"),
            ("Contig", "@seq_id"),
            ("Length", "@length"),
            ("Coverage (%)", '@coverage'),
            ("Cluster", "@cluster"),
        ]
    )
    figure.add_tools(hover)

    figure.xaxis.axis_label = x_label
    figure.yaxis.axis_label = y_label

    return figure, datasource


def output_html_pca(source, pca_plot_main, pca_plot_sub, file_name):
    LOG.info("Ouput PCA plot to %s", file_name)

    columns = [
        TableColumn(field='seq_id', title='Contig'),
        TableColumn(field='lineage', title='Lineage'),
        TableColumn(field='length', title='Length'),
        TableColumn(field='coverage', title='Coverage (%)'),
        TableColumn(field='cluster', title='Cluster'),
        TableColumn(field='target_taxon', title='Target Taxon'),
    ]
    data_table = DataTable(
        source=source,
        columns=columns,
        width=1024,
        height=600
    )

    div_change = Div(
        sizing_mode="stretch_both",
        text="",
        width=200,
        height=800
    )
    source.callback = CustomJS(args=dict(ti=div_change, dt=source), code="""
            var inds = cb_obj.selected['1d'].indices;
            var d1 = cb_obj.data;

            var contigs = [];

            for (i = 0; i < inds.length; i++){
                contigs.push(d1['seq_id'][inds[i]])
            }

            contigs = Array.from(new Set(contigs));
            contigs.sort();
            ti.text = contigs.join("<br>");
        """)

    grid_plot = gridplot(
        [[pca_plot_main], pca_plot_sub, [data_table], [div_change]],
        merge_tools=False,
        toolbar_location='left',
        sizing_mode="fixed",
    )
    # grid_plot.sizing_mode = 'scale_height'

    html = file_html(grid_plot, CDN, "PCA Contigs")

    open(file_name, 'w').write(html)


def prepare_gc_data(sequences, bins):
    LOG.info("Calculate GC content")

    gc_content = {
        seq_id: mgkit.utils.sequence.sequence_gc_content(seq) * 100
        for seq_id, seq in sequences.items()
    }
    gc_content = pandas.Series(gc_content)
    dataframe = pandas.DataFrame(
        {
            'bin': pandas.cut(gc_content, bins=bins),
            'gc_content': gc_content
        }
    ).sort_values(by='bin')

    gc_counts = dataframe.groupby(by='bin').count()
    gc_counts['bin'] = [str(x) for x in gc_counts.index]
    gc_counts['left'] = pandas.cut(gc_content, bins=bins, retbins=True)[1][:-1]
    gc_counts['avg_gc'] = dataframe.groupby(by='bin').mean()
    gc_counts['contigs'] = dataframe.groupby(by='bin').apply(
        lambda x: '<br>'.join(x.index)
    )

    gc_counts.index = range(len(gc_counts))

    return gc_counts


def output_html_gc_content(gc_counts, file_name):
    LOG.info("Ouput GC plot to %s", file_name)

    datasource = ColumnDataSource(gc_counts)

    div_change = Div(
        sizing_mode="stretch_both", text="", width=800, height=200
    )
    callback = CustomJS(
        args=dict(div=div_change, ds=datasource),
        code="""
        var indices = cb_data.index['1d'].indices;
        var div_content = [];
        var datas = ds.data;
        for (i=0; i < indices.length; i++) {
            div_content.push(datas['contigs'][indices[i]]);
        }
        div_content.sort();
        if (div_content.length > 0) {
            div.text = div_content.join("<br>");
        }
        """
    )

    hover = HoverTool(
        tooltips=[
            ('Average GC %', '@avg_gc'),
            ('Number of Contigs', '@gc_content'),
        ],
        callback=callback
    )

    figure = bokeh.plotting.figure(
        width=800,
        height=300,
        tools=','.join(
            [
                "pan",
                "box_zoom",
                "save",
                "reset",
                "tap",
                "lasso_select",
                "box_select",
                # "resize", # doesn't work with the layout
                "wheel_zoom"
            ]
        ),
    )

    width = (
        gc_counts.left[1:].values - gc_counts.left[:-1].values
    ).min() * 0.8

    figure.vbar(
        x='left',
        width=width,
        top='gc_content',
        source=datasource,
    )
    figure.xaxis.axis_label = 'GC %'
    figure.yaxis.axis_label = 'Number of Contigs'

    figure.add_tools(hover)

    grid = column(
        figure,
        div_change,
        sizing_mode='scale_width',
    )

    html = file_html(grid, CDN, "GC% Contigs")

    open(file_name, 'w').write(html)



try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

PACKAGE_NAME = 'scm'

requirements = [
    # Main library for LCA/GFF
    'mgkit>=0.4.1',
    # Matrix manipulation
    'numpy>=1.11.0',
    'pandas>=0.20.0',
    'scipy>=0.19',  # currently not imported directly
    # HDF5 manipulation, required by pandas
    'tables>=3.4.2',
    # Clustering/PCA
    'scikit-learn>=0.18',
    # Smaller utilities
    'setuptools>=34.0.2',
    'msgpack-python>=0.4.8',
    'progressbar2>=3.30.2',
    'Distance>=0.1.3',
    # Plotting libraries
    'bokeh>=0.12.9',
    'seaborn>=0.7.1',
    'matplotlib>=2.0',
]

setup(
    name="single-cell-m",
    author='Francesco Rubino',
    author_email='rubino.francesco@gmail.com',
    description='Single Cell M',
    long_description=open('README.rst').read(),
    version=open('VERSION').read(),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            '{0} = {0}.main:main'.format(PACKAGE_NAME),
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['scripts', 'docs', 'test-data']),
    scripts=[
        'scripts/scm-download-taxa.sh'
    ],
)
